/* Copyright (c) 2006, Catalin Patulea <cat@vv.carleton.ca> All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 * 
 *     * Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.
 * 
 *     * Neither the name of Catalin Patulea nor the names of its contributors may 
 *     be used to endorse or promote products derived from this software without 
 *     specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "stdafx.h"
#include "Display.h"
#include "Util.h"
#include "resource.h"

namespace Display {

// SSD0303 State
BOOL On = FALSE;				// Display On/Off
unsigned char Column = 0;		// Column Address Register
unsigned char Contrast = 0x80;	// Constrast Control Register
BOOL SegmentRemap = FALSE;		// "column address 131 is mapped to SEG0"
unsigned char StartLine = 0;	// Display Start Line
unsigned char Offset = 0;		// Vertical scroll/Display Offset
unsigned char MuxRatio = 0x3F;	// Multiplex Ratio
BOOL Flood = 0;					// Entire Display On/Off
BOOL Invert = 0;				// Inverted On/Off
unsigned char Page = 0;			// GDRAM Page (0-7)
BOOL ScanRemapped = 0;			// Output Scan Direction (COM N-1 to COM 0 when TRUE)
unsigned char Clock = 0x70;		// Display Clock Divide Ratio/Oscillator Frequency
unsigned char ColorMode = 0x00;	// area color mode on/off & low power display mode
unsigned char Precharge = 0x22;	// Pre-charge period
BOOL COMNonsequential = 1;		// Alternative COM pin configuration when TRUE
unsigned char VCOMDeselect = 0x00; // VCOM Deselect Level
BOOL DCDC = TRUE;				// DC-DC will be turned on when display on

// UI framebuffer data
static const RGBQUAD Black = {0, 0, 0, 0};
static const RGBQUAD Orange = {0, 0x80, 0xFF, 0};

#include <pshpack1.h>
static struct {
	BITMAPINFOHEADER bmiHeader;
	RGBQUAD bmiColors[2];
	BYTE bmiData[16][12]; // 16 rows of 12*8 columns
} bi;
#include <poppack.h>

void DialogUpdate(void);
void DialogKill(AGSIDLGD *pM);
void DialogDisp(AGSIMENU *pM);

// must not use 'const' here !
//                     iOpen  Hwnd   Dlg Proc.  Rect: -1 := default    Update       Kill
static AGSIDLGD Dlg = {    0,   NULL,      NULL,  { -1, -1, -1, -1, },  &DialogUpdate,  &DialogKill};
static AGSIMENU Menu =
	{  1,   "&SSD0303"        , &DialogDisp,  0,   IDD_DISPLAY,      &Dlg};

// Largest command block is 5 bytes
static unsigned char Buffer[5];

static unsigned int BufferIndex = 0;

#define FIRST(x,n)	((x) & (unsigned char) (0xFF00 >> (n)))
#define LAST(x,n)	((x) & (unsigned char)~(0x00FF << (n)))

// Schedules the display window for redrawing ASAP
static void InvalidateDisplay(void) {
	// Invalidate display window
	if (Dlg.hw) {
		InvalidateRect(GetDlgItem(Dlg.hw, IDC_DISPLAY), NULL, TRUE);
	}
}

static void HandleCommand(unsigned char Value) {
	// Don't overflow
	if (BufferIndex >= sizeof(Buffer)) {
		return;
	}

	Buffer[BufferIndex++] = Value;

	// Process command
	unsigned char Command = Buffer[0];
	unsigned int CommandLength = 1; // Most command are only 1 byte
	if (FIRST(Command, 7) == 0xAE) {
		On = LAST(Command, 1);
	} else if (FIRST(Command, 4) == 0x00) {
		Column = (Column & 0xF0) | LAST(Command, 4);
	} else if (FIRST(Command, 4) == 0x10) {
		Column = (Column & 0x0F) | (LAST(Command, 4) << 4);
	} else if (FIRST(Command, 8) == 0x81) {
		CommandLength = 2;
		if (BufferIndex == CommandLength) {
			Contrast = Buffer[1];
		}
	} else if (FIRST(Command, 7) == 0xA0) {
		SegmentRemap = LAST(Command, 1);
	} else if (FIRST(Command, 2) == 0x40) {
		StartLine = LAST(Command, 6);
	} else if (FIRST(Command, 8) == 0xD3) {
		CommandLength = 2;
		if (BufferIndex == CommandLength) {
			Offset = LAST(Buffer[1], 6);
		}
	} else if (FIRST(Command, 8) == 0xA8) {
		CommandLength = 2;
		if (BufferIndex == CommandLength) {
			MuxRatio = LAST(Buffer[1], 6);
		}
	} else if (FIRST(Command, 7) == 0xA4) {
		Flood = LAST(Command, 1);
	} else if (FIRST(Command, 7) == 0xA6) {
		Invert = LAST(Command, 1);
	} else if (FIRST(Command, 4) == 0xB0) {
		Page = LAST(Command, 3);
	} else if (FIRST(Command, 4) == 0xC0) {
		ScanRemapped = LAST(Command >> 3, 1);
	} else if (FIRST(Command, 8) == 0xD5) {
		CommandLength = 2;
		if (BufferIndex == CommandLength) {
			Clock = Buffer[1];
		}
	} else if (FIRST(Command, 8) == 0xD8) {
		CommandLength = 2;
		if (BufferIndex == CommandLength) {
			ColorMode = Buffer[1] & 0x35;
		}
	} else if (FIRST(Command, 8) == 0xD9) {
		CommandLength = 2;
		if (BufferIndex == CommandLength) {
			Precharge = Buffer[1];
		}
	} else if (FIRST(Command, 8) == 0xDA) {
		CommandLength = 2;
		if (BufferIndex == CommandLength && (Buffer[1] & 0xEF) == 0x02) {
			COMNonsequential = LAST(Buffer[1] >> 4, 1);
		}
	} else if (FIRST(Command, 8) == 0xDB) {
		CommandLength = 2;
		if (BufferIndex == CommandLength) {
			VCOMDeselect = LAST(Buffer[1], 6);
		}
	} else if (FIRST(Command, 8) == 0xAD) {
		CommandLength = 2;
		if (BufferIndex == CommandLength && FIRST(Buffer[1], 7) == 0x8A) {
			DCDC = LAST(Buffer[1], 1);
		}
	} else {
		// TODO: Display warning about unsupported command.
	}

	// Clear buffer and perform some updating if command is complete
	if (BufferIndex == CommandLength) {
		BufferIndex = 0;

		// Refresh the palette
		if (Flood) {
			bi.bmiColors[0] = Orange;
			bi.bmiColors[1] = Orange;
		} else if (Invert) {
			bi.bmiColors[0] = Orange;
			bi.bmiColors[1] = Black;
		} else {
			bi.bmiColors[0] = Black;
			bi.bmiColors[1] = Orange;
		}

		InvalidateDisplay();
	}
}

static void HandleData(unsigned char Value) {
	// Last 96 columns are connected to display
	BYTE RealColumn = Column - 0x24;

	// Write to UI framebuffer
	for (unsigned int i = 0; i < 8; i++) {
		BYTE& Cell = bi.bmiData[(Page << 3) + i][RealColumn >> 3];
		unsigned int Position = RealColumn & 7;

		Cell = (Cell & ~((BYTE)0x80 >> Position)) | ((Value & 1) << (7 - Position));
		Value >>= 1;
	}

	Column++;

	InvalidateDisplay();
}

void HandleTransfer(std::deque<unsigned char>& Queue) {
	// Reset command buffer
	BufferIndex = 0;

	while (Queue.size() != 0) {
		unsigned char Value = Queue.front();
		Queue.pop_front();

		// Check Co bit
		if (Value & 0x80) {
			if (Queue.size() == 0) {
				// TODO: Display warning about malformed command.
				return;
			}

			HandleCommand(Queue.front());
			Queue.pop_front();
		} else {
			// Break out and process remainder as data only
			break;
		}
	}

	while (Queue.size() != 0) {
		HandleData(Queue.front());
		Queue.pop_front();
	}
}

// UI
LRESULT CALLBACK DisplayProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_PAINT: {
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(hWnd, &ps);
		RECT rc;
		GetClientRect(hWnd, &rc);

		if (On && DCDC) {
			StretchDIBits(hdc,
				rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top,
				0, bi.bmiHeader.biHeight, bi.bmiHeader.biWidth, -bi.bmiHeader.biHeight,
				bi.bmiData, (const BITMAPINFO *)&bi,
				DIB_RGB_COLORS, SRCCOPY);
		} else {
			FillRect(hdc, &rc, (HBRUSH)(COLOR_BTNFACE+1));
		}

		EndPaint(hWnd, &ps);
		return 0;
	}
	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
}

#define DISPLAY_CLASS "SSD0303DisplayClass"

BOOL Init(void) {
	// Set up monochrome bitmap framebuffer
	ZeroMemory(&bi, sizeof(bi));
	bi.bmiHeader.biSize = sizeof(bi.bmiHeader);
	bi.bmiHeader.biWidth = 96;
	bi.bmiHeader.biHeight = 16;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = 1;
	bi.bmiHeader.biCompression = BI_RGB;
	bi.bmiHeader.biSizeImage = sizeof(bi.bmiData);
	bi.bmiHeader.biClrUsed = 2;

	bi.bmiColors[0] = Black;
	bi.bmiColors[1] = Orange;

	// Create display window class
	static WNDCLASSEX wcx = {sizeof(wcx), CS_PARENTDC, DisplayProc, 0, 0, _instance,
		NULL, LoadCursor(NULL, IDC_ARROW), NULL, NULL, DISPLAY_CLASS, NULL};
	return RegisterClassEx(&wcx);
}

BOOL Terminate(void) {
	return UnregisterClass(DISPLAY_CLASS, _instance);
}

static BOOL CALLBACK DialogProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
	case WM_INITDIALOG: {
		RECT rc = {0, 0, 2 * bi.bmiHeader.biWidth, 2 * bi.bmiHeader.biHeight};

		// Adjust size, but leave at same position
		AdjustWindowRectEx(&rc, GetWindowLong(GetDlgItem(hWnd, IDC_DISPLAY), GWL_STYLE),
			FALSE, GetWindowLong(GetDlgItem(hWnd, IDC_DISPLAY), GWL_EXSTYLE));
		SetWindowPos(GetDlgItem(hWnd, IDC_DISPLAY), NULL, 0, 0,
			rc.right - rc.left, rc.bottom - rc.top,
			SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOREDRAW |
			SWP_NOZORDER);
	}
	case WM_CLOSE:
		DialogKill(&Dlg);
		return TRUE;
	default:
		return FALSE;
	};
}

static void DialogUpdate(void)  {
	UpdateWindow(GetDlgItem(Dlg.hw, IDC_DISPLAY));
}

static void DialogKill(AGSIDLGD *pDlg) {
	DestroyWindow(pDlg->hw);
	pDlg->hw = NULL;
}

static void DialogDisp(AGSIMENU *pM) {
	// Create the dialog (if not already created)
	if (pM->pDlg->hw == NULL) {
		pM->pDlg->hw = CreateDialog(_instance, MAKEINTRESOURCE(pM->nDlgId),
			NULL, DialogProc);
	} else {
		DialogKill(pM->pDlg);
	}
}

BOOL DefineAllMenuEntries(void) {
	return Agsi::DefineMenuItem(&Menu);
}

};
